


MOVFM1      =  $BBA2    ; Load FP value from memory into FAC1
MOVFM2      =  $BA8C    ; Load value from memory into FAC2 (ARG)
MOVMF       =  $BBD4    ; Store FAC1 into memory address pointed to by A/Y (lo/hi)
FADDT       =  $B86A
FSUBT       =  $B853
FMULT       =  $BA2B
FDIVT       =  $BB12
FPOW        =  $BF7B    ; FAC = ARG ^ FAC
FINT        =  $BCCC    ; FAC = INT(FAC)    (truncates to integer)

FOUT        =  $BDDD    ; Convert FAC to string, leave string address in A/Y (lo/hi)
STRVAL      =  $B7B0    ; Convert string to number in FAC.  $22/23 points to string, length in .A

STROUT      =  $AB1E    ; Output null-terminated string pointed to by A/Y (lo/hi)
CHROUT      =  $F1CA
FACINT      =  $BC9B    ; Convert FAC to 4-byte integer (signed or unsigned).  Stores at $62 (FAC +1)
;CMPFLT      =  $BC5B    ; Compare ARG ? FAC (where ? is the comparison operator in location CMPCODE)
;FNABS       =  $BC58
FNEXP       = $BFED     ; FAC = EXP(FAC)
STRIN       = $A560     ; Allows typing at cursor position, reads what you typed from the screen, into buffer at $200

FAC         =  $61
ARG         =  $69
CMPCODE     =  $4D

InitIP      =  $49

            *= $1000


; Use A/X as low/high byte of code to start executing.
;InterpEnter LDA InitIP           ; Initialize IP to requested starting point
InterpEnter LDA #$00
            STA IP
;            LDA InitIP +1
            LDA #$20
            STA IP +1            ; Byte-code location hard-coded to $2000
            
            TSX
            STX   SPSave         ; Save SP so we can restore it before returning to BASIC   
            
            LDY   #>VersionStr   ; Report version string
            LDA   #<VersionStr
            JSR   STROUT
            
            JMP InterpLoop


VersionStr  .TEXT    "Stack-based interpreter v0.3"
            .BYTE    13, 0
SPSave      .BYTE    0              ; save stack pointer when we start, restore it when we return to BASIC
            
; ----------------------
; NOP handler 
; ----------------------
            ; Do nothing...
OpNop       JMP InterpLoop


; ----------------------
; Add handler 
; ----------------------
OpAdd       JSR PopFac        ; Pop FAC and ARG from stack
            JSR PopArg
            LDA FAC
            JSR FADDT         ; Add
            JSR PushFac       ; Push FAC onto stack
            JMP InterpLoop


; ----------------------
; Subtract handler 
; ----------------------
OpSubtract  JSR PopFac        ; Pop FAC and ARG from stack
            JSR PopArg
            LDA FAC
            JSR FSUBT         ; FAC = ARG - FAC
            JSR PushFac       ; Push FAC onto stack
            JMP InterpLoop


; ----------------------
; Multiply handler 
; ----------------------
OpMultiply  JSR PopFac        ; Pop FAC and ARG from stack
            JSR PopArg
            LDA FAC
            JSR FMULT         ; Add
            JSR PushFac       ; Push FAC onto stack
            JMP InterpLoop


; ----------------------
; Divide handler 
; ----------------------
OpDivide    JSR PopFac        ; Pop FAC and ARG from stack
            JSR PopArg
            LDA FAC
            JSR FDIVT         ; Add
            JSR PushFac       ; Push FAC onto stack
            JMP InterpLoop


; ----------------------
; Concatenate handler 
; ----------------------
            ; Not yet implemented
OpConcatStr LDY   #>ConcatStrNI
            LDA   #<ConcatStrNI
            JSR   STROUT
            JMP BackToBasic
ConcatStrNI .TEXT "Opcode 05 (Concatenate string) not yet implemented"
            .BYTE 13,0

; ----------------------
; Drop handler 
; ----------------------
            ; Not yet implemented
OpDrop      LDY   #>DropNI
            LDA   #<DropNI
            JSR   STROUT
            JMP BackToBasic
DropNI      .TEXT "Opcode 06 (Drop) not yet implemented"
            .BYTE 13,0


            
            
                        
InterpLoop  JSR CodeFetch
            CLC
            ASL A
            CMP   #JumpTblSz
            BCS   OpcodeNI
            TAY
            LDA   JumpTable+1,Y
            PHA
            LDA   JumpTable,Y
            PHA
            RTS      ; Used as a jump!
            
            ; If unknown opcode, report it and return to BASIC
OpcodeNI    LSR      ; Convert back to opcode (was x2 for table lookup)
            PHA
            LSR
            LSR
            LSR
            LSR
            CMP   #$0A
            BCC   _noAlpha1
            ADC   #$05        ; difference between "9"+1 and "A", less one (carry flag set)
_noAlpha1   ADC   #$30
            JSR   CHROUT
            PLA
            AND   #$0F
            CMP   #$0A
            BCC   _noAlpha2
            ADC   #$05        ; difference between "9"+1 and "A", less one (carry flag set)
_noAlpha2   ADC   #$30
            JSR   CHROUT

            LDY   #>OpcodeNI1
            LDA   #<OpcodeNI1
            JSR   STROUT
            JMP   BackToBasic
OpcodeNI1   .TEXT ": Unrecognized Opcode.  Aborting! "
            .BYTE 13, 0
            
      ;NOP = 0
      ;Add = 1
      ;Subtract = 2
      ;Multiply = 3
      ;Divide = 4
      ;Concatenate = 5
      ;Drop = 6
      ;Fetch = 7
      ;Store = 8
      ;Swap = 9
      ;Input = A
      ;PrintNum = B
      ;PrintString = C
      ;NewLine = D
      ;LiteralData = E
      ;EndStatement = F
      ;EndProgram = 10
      ;Compare = 11
      ;JumpRel = 12
      ;JumpAbs = 13
      ;BranchFalse = 14

; ----------------------
; Fetch handler 
;   Takes an address off of the stack (as a real, doh!) and pushes
;   the real number stored in memory at that address.
; ----------------------
OpFetch     JSR   PopFac      ; Treat top item as an address...
            JSR   FACINT      ; Convert to integer format (BIG-Endian for some reason)
            LDA   FAC+4
            LDY   FAC+3       ; A/Y = address (lo/hi)
            JSR   MOVFM1      ; Transfer value into FAC
            JSR   PushFac     ; Push the value onto the stack
            JMP   InterpLoop


; ----------------------
; Store handler 
; ----------------------
            ; Not yet implemented
OpStore     JSR   PopFac      ; Treat top item as an address...
            JSR   FACINT      ; Convert to integer format (BIG-Endian for some reason)
            LDA   FAC+4       ; A/Y = address (lo/hi)
            LDY   FAC+3       
            STA   ARG+1       ; remember address...
            STY   ARG+2

            JSR   PopFac      ; Now pull value off of stack
            LDX   ARG+1       ; Retrieve address into x/y (lo/hi)
            LDY   ARG+2
            JSR   MOVMF       ; Store value in memory at that address
            JMP InterpLoop


; ----------------------
; Input handler 
; ----------------------
            ; Not yet implemented
OpInput     LDY   #>InputNI
            LDA   #<InputNI
            JSR   STROUT
            JMP BackToBasic
InputNI     .TEXT "Opcode 0A (Input) not yet implemented"
            .BYTE 13,0


; ----------------------
; PrintNum handler 
; ----------------------
OpPrintNum  JSR PopFac
            JSR FOUT
            JSR STROUT
            JMP InterpLoop


; ----------------------
; PrintString handler 
; ----------------------
OpPrintStr  = *
            PLA                  ; String length
            BEQ   _blank
            STA   $FF
_loop       PLA                  ; Output the next char
            JSR   CHROUT
            DEC   $FF
            BNE   _loop
_blank      JMP   InterpLoop


;OpPrintStr  LDY   #>PrintStrNI
;            LDA   #<PrintStrNI
;            JSR   STROUT
;            JMP BackToBasic
;PrintStrNI  .TEXT "Opcode 0C (Print String) not yet implemented"
;            .BYTE 13,0


; ----------------------
; NewLine handler 
; ----------------------
OpNewline   LDA #$0D
            JSR CHROUT
            JMP InterpLoop

            
; ----------------------
; EndStatement handler 
; ----------------------
            ; Not yet implemented
OpEndStmt   JMP BackToBasic


; ----------------------
; EndProgram handler 
; ----------------------
            ; We're done, leave interpreter
OpEndProg   JMP BackToBasic



; ----------------------
; LiteralData handler 
; ----------------------
OpLiteral   JSR CodeFetch     ; Get next byte from code stream into reg A
            TAX            
LiteralLp   JSR CodeFetch
            PHA
            DEX
            BNE LiteralLp
            JMP InterpLoop


; ----------------------
; Swap handler 
;
; The top stack item trades places with the one below it.
; Both items are 5-byte values.  Uses $FB-FE for indirect addressing.
; ----------------------
OpSwap      TSX
            TXA                  ; Points 1 below first byte of top item (LSB)
            STA   $FB
            LDY   #$01           ; MSB of stack page
            STY   $FC
            CLC
            ADC   #$05           ; Points just before first byte of next item down
            STA   $FD
            STY   $FE            ; same MSB

            LDY   #5             ; # bytes to swap, also starting index
_loop       LDA   ($FB),Y
            TAX                  ; Use .X as temp
            LDA   ($FD),Y
            STA   ($FB),Y
            TXA
            STA   ($FD),Y
            
            DEY
            BNE   _loop
            JMP   InterpLoop


; ----------------------
; Compare handler 
; ----------------------
OpCompare   = *
            ; Read test type from code stream, store for later use
            JSR   CodeFetch
            STA   CMPCODE
            
            ; Put top 2 values on stack into FAC and ARG
            JSR   PopFac         ; b
            JSR   PopArg         ; a
            
            ; Compare by subtracting second argument from first.
            JSR   FSUBT
            
            ; Result:  0=(a==b), negative=(a<b), positive=(a>b)
            LDA   #2             ; Code for equal
            LDX   FAC +0         ; Exponent (exp=0 means a==b)
            BEQ   _cmpdone
            LDA   #4             ; Code for <
            LDX   FAC +5         ; Sign bit (negative if bit 7 set, meaning a<b)
            BMI   _cmpdone
            LDA   #1             ; Only remaning possibility...  a>b

            ; if result of comparison matches any bit in test mask, result is true!
_cmpdone    AND   CMPCODE
            BEQ   _pushres       ; FALSE = 0 (FP: 00 00 00 00 00)
            LDA   #$81           ; TRUE  = 1 (FP: 81 00 00 00 00)
_pushres    TAX                  ; last byte to push on stack
            LDA   #$00
            PHA
            PHA
            PHA
            PHA
            TXA                  ; Get back exp byte (only difference between true/false)
            PHA
            JMP   InterpLoop


;OpCompare   LDY   #>CompareNI
;            LDA   #<CompareNI
;            JSR   STROUT
;            JMP BackToBasic
;CompareNI   .TEXT "Opcode 11 (Compare) not yet implemented"
;            .BYTE 13,0



; ----------------------
; JumpRel handler 
; ----------------------
OpJumpRel   JSR   CodeFetch         ; Get lsb of offset
            PHA
            JSR   CodeFetch         ; msb of offset
            TAX
            PLA
            
            ; Add offset to program counter
            CLC
            ADC   IP
            STA   IP
            TXA
            ADC   IP+1
            STA   IP+1
            
            ; Begin fetch code from new location!
            JMP   InterpLoop

;OpJumpRel   LDY   #>JumpRelNI
;            LDA   #<JumpRelNI
;            JSR   STROUT
;            JMP BackToBasic
;JumpRelNI   .TEXT "Opcode 12 (Jump Relative) not implemented!"
;            .BYTE 13,0


; ----------------------
; JumpAbs handler 
; ----------------------
OpJumpAbs   LDY   #>JumpAbsNI
            LDA   #<JumpAbsNI
            JSR   STROUT
            JMP BackToBasic
JumpAbsNI   .TEXT "Opcode 13 (Jump Absolute) not implemented!"
            .BYTE 13,0


; ----------------------
; BranchFalse handler 
; ----------------------
OpBrFalse   JSR   PopFac
            LDA   FAC
            BEQ   OpJumpRel      ; if value = 0 (false), jump around some code
            
            ; non-zero value (true):  discard optional branch address and "fall through"
            JSR   CodeFetch      ; toss 2 offset bytes
            JSR   CodeFetch
            JMP   InterpLoop


; ----------------------
; Exponent handler
; ----------------------
OpExponent  JSR   PopFac
            JSR   FNEXP
            JSR   PushFac
            
            JMP   InterpLoop


; ----------------------
; Input Number handler
; ----------------------
OpInputNum  JSR   STRIN       ; Get string into buffer at $200
            
            ; Point to buffer...
            LDA   #$00        
            STA   $22
            LDA   #$02
            STA   $23
            
            ; Find the length of the string...
            LDY   #$FF
_lenLoop    INY
            LDA   ($22),Y
            BNE   _lenLoop
            TYA               ; Length now in .A
            
            JSR   STRVAL      ; Convert string to number in FAC
            JSR   PushFac
            JMP   InterpLoop


; ----------------------
; Input String handler
; ----------------------
;OpExponent  = *
OpInputStr  LDY   #>InputStrNI
            LDA   #<InputStrNI
            JSR   STROUT
            JMP BackToBasic
InputStrNI  .TEXT "Opcode 16 (Input String) not implemented!"
            .BYTE 13,0

            JMP   InterpLoop


; ----------------------
; Power handler (a ^ b)
; ----------------------
OpPower     JSR   PopFac
            JSR   PopArg
            LDA   FAC
            JSR   FPOW
            JSR   PushFac
            JMP   InterpLoop


; ----------------------
; Truncate handler
; ----------------------
OpTrunc     JSR   PopFac
            JSR   FINT
            JSR   PushFac
            JMP   InterpLoop



; ----------------------
; CodeFetch sub
; ----------------------
CodeFetch   LDA $FFFF            ; Self-modifying code, address to be filled in later.
IP          =   * -2
            INC IP               ; Increment IP to next address.
            BNE CodeFetch2
            INC IP +1
CodeFetch2  RTS
            

; ----------------------
; PopArg sub
; ----------------------
PopArg      TSX               ; Put address of top value on stack into .A/.Y (lo/hi)
            TXA
            CLC
            ADC   #3          ; Account for return address and SP pointing to next free space
            LDY   #$01        ; .Y is MSB of address (stack page)
            JSR   MOVFM2      ; Move FP # from memory to FAC2 (ARG)
            
            PLA               ; Preserve return address in x/y
            TAX
            PLA
            TAY
            PLA               ; remove FP value from stack
            PLA
            PLA
            PLA
            PLA
            TYA               ; Restore return address
            PHA
            TXA
            PHA
            RTS
            

; ----------------------
; PopFac sub
; ----------------------
PopFac      TSX               ; Put address of top value on stack into .A/.Y (lo/hi)
            TXA
            CLC
            ADC   #3          ; Account for return address and SP pointing to next free space
            LDY   #$01        ; .Y is MSB of address (stack page)
            JSR   MOVFM1      ; Move FP # from memory to FAC1
            
            PLA               ; Preserve return address in x/y
            TAX
            PLA
            TAY
            PLA               ; remove FP value from stack
            PLA
            PLA
            PLA
            PLA
            TYA               ; Restore return address
            PHA
            TXA
            PHA
            RTS
            

; ----------------------
; PushFac sub
; ----------------------
PushFac     pla               ; Take return address off of the stack
            tax
            pla
            tay
            pha               ; Push 5-byte place-holder onto stack
            pha
            pha
            pha
            pha
            tya               ; Put return address back on top of stack
            pha
            txa
            pha
            
            tsx
            txa
            clc
            adc   #3          ; Calc LSB of value "under" return address
            tax
            ldy   #$01        ; .Y holds MSB (stack page)
            jsr   MOVMF
            rts
            

; ----------------------
; BackToBasic jump point
; ----------------------
BackToBasic LDX   SPSave      ; Restore stack pointer to what it was when the interpreter started,
            TXS               ; to avoid problems if the code left junk on the stack
            RTS            

; ----------------------
; sub
; ----------------------
            

; ----------------------
; sub
; ----------------------
            

; ----------------------
; sub
; ----------------------


JumpTable   .WORD OpNop-1, OpAdd-1, OpSubtract-1, OpMultiply-1, OpDivide-1, OpConcatStr-1, OpDrop-1, OpFetch-1, OpStore-1, OpSwap-1
            .WORD OpInput-1, OpPrintNum-1, OpPrintStr-1, OpNewline-1, OpLiteral-1, OpEndStmt-1, OpEndProg-1, OpCompare-1, OpJumpRel-1,
            .WORD OpJumpAbs-1, OpBrFalse-1, OpExponent-1, OpInputStr-1, OpInputNum-1, OpPower-1, OpTrunc-1
JumpTblSz   =  * - JumpTable


